@@ -151,12 +151,11 @@
 @class	NSThread;
 @class	NSUserDefaults;
 
-/**
- * This protocol is implemented by a delegate of a WebServer instance
+/** This protocol is implemented by a delegate of a WebServer instance
  * in order to allow the delegate to process requests which arrive
  * at the server.
  */
-@protocol	WebServerDelegate
+@protocol	WebServerDelegate <NSObject>
 
 /**
  * Process the HTTP request whose headers and data are provided in
@@ -248,13 +247,20 @@
 		    for: (WebServer*)http;
 @end
 
-
 /** This is an informal protocol documenting optional methods which will
  * be used if implemented by the delegate.
  */
 @interface	NSObject(WebServerDelegate)
 
-/* If your delegate implements this method, it will be called before the
+/** Informs the handler (if any) associated with the response that the
+ * server has written the response to the network.  The timeInterval
+ * is measured between the point when the server started reading the
+ * request and the point at which it finished writing the response.
+ */
+- (void) completedResponse: (WebServerResponse*)response
+		  duration: (NSTimeInterval)timeInterval;
+
+/** If your delegate implements this method, it will be called before the
  * first call to handle a request (ie before -preProcessRequest:response:for:
  * or -processRequest:response:for:) to provide the delegate with the
  * request header information and allow it to decide whether the request
@@ -430,6 +436,7 @@
   NSLock		*_lock;
   IOThread		*_ioMain;
   NSMutableArray	*_ioThreads;
+  NSMutableDictionary	*_blocked;
   GSThreadPool		*_pool;
   WebServerConfig	*_conf;
   id		        UNUSED_QUAL *_unused1;
@@ -466,6 +473,7 @@
   NSMutableDictionary   *_incrementalDataMap;
   NSUInteger            _strictTransportSecurity;
   NSString              *_frameOptions;
+  NSTimeInterval        _authBlock;
   void			*_reserved;
 }
 
@@ -586,6 +594,14 @@
  */
 - (NSString*) address;
 
+/** If greater than zero, the returned value is the number of seconds for
+ * which the server should block subsequent requests from the offending
+ * address, otherwise (returned value is zero) bocking is not performed
+ * when an authentication attem,pt fails.<br />
+ * Blocked requests will get a 429 response.
+ */
+- (NSTimeInterval) blockOnAuthenticationFailure;
+
 /**
  * Instructs the server that the connection handlind the current request
  * should be closed once the response has been sent back to the client.
@@ -700,7 +716,7 @@
 - (BOOL) isSecure;
 
 /**
- * Returns YES if the server is running behiond a secure proxy inside
+ * Returns YES if the server is running behind a secure proxy inside
  * a demilitarised zone (DMZ).
  */
 - (BOOL) isTrusted;
@@ -806,6 +822,15 @@
 	    fromTemplate: (NSString*)aPath
 		   using: (NSDictionary*)map;
 
+/** Sets the time for which requests from the same host should be blocked
+ * if a request from the host attempts to authenticate and fails.<br />
+ * The default is 1 second but setting a value of zero or less turns this
+ * feature off (sets the blocking interval to zero).<br />
+ * The -[WebServerResponse block:] method may be used to
+ * set a different timeout in response to a particular request.
+ */
+- (void) setBlockOnAuthenticationFailure: (NSTimeInterval)ti;
+
 /**
  * Sets the time after which an idle connection should be shut down.<br />
  * Default is 30.0
@@ -835,7 +860,7 @@
  * and the connection between the proxy and the server is also secure.<br />
  * If the dictionary also contains <code>HSTS</code> then this must be a
  * non-negative integer value specifying the number of seconds to set
- * in the Strict-Transport-Security header (defaults to 7 days).<br />
+ * in the Strict-Transport-Security header (defaults to 1 year).<br />
  * The <em>secure</em> dictionary may also contain other dictionaries
  * keyed on IP addresses, and if the address that an incoming connection
  * arrived on matches the key of a dictionary, that dictionary is used
@@ -847,8 +872,12 @@
  * Typically a failure will be due to an invalid address or port being
  * specified ...  a port may not already be in use and may not be in the
  * range up to 1024 (unless running as the super-user).<br />
- * Call this with a nil/empty port argument to shut the server down as soon as
- * all current connections are closed (and refuse new incoming connections).
+ * Call this with a nil/empty port argument to shut the server down as soon
+ * as all current connections are closed (and refuse new incoming
+ * connections).<br />
+ * NB. Changing of this configuration must only occur in the master thread
+ * so this method, if called from another thread, will need to perform some
+ * internal operations in the master thread before it returns.
  */
 - (BOOL) setAddress: (NSString*)anAddress
 	       port: (NSString*)aPort
@@ -976,8 +1005,10 @@
  * Sets the maximum number of connections in each I/O thread which are
  * to be kept in a 'keepalive' state waiting for a new request once a
  * request completes.<br />
- * The permitted range is currently from 1 to 1000, with settings outside
- * that range being treated as the default value (100).
+ * The permitted range is currently from 0 to 1000, with settings being
+ * limited to that range.  The default value is 0, which means that the
+ * number of idle connections per thread is unlimited (though the total
+ * number of connections and number per host is still constrained).
  */
 - (void) setMaxKeepalives: (NSUInteger)max;
 
@@ -991,7 +1022,8 @@
 
 /** Sets the HTTP methods which may be used by the server.<br />
  * Any incoming request using a method not listed in the permitted set is
- * rejected with an HTTP 501 response.<br />
+ * rejected with an HTTP 405 response and an Allow header saying which
+ * methods ARE allowed.<br />
  * The default set contains only the GET and POST methods.
  */
 - (void) setPermittedMethods: (NSSet*)s;
@@ -1019,7 +1051,7 @@
  * to a request on a secure connection (including via a secure proxy).<br />
  * The Strict-Transport-Security header is automatically set in the response
  * to any incoming request (but code handling the request may alter that).<br />
- * The default setting is 7 days (604800 seconds), while a setting of zero
+ * The default setting is 1 year (31536000 seconds), while a setting of zero
  * turns off HSTS.
  */
 - (void) setStrictTransportSecurity: (NSUInteger)seconds;
@@ -1119,6 +1151,10 @@
 		   into: (NSMutableString*)result
 		  depth: (NSUInteger)depth;
 
+/** Returns the thread pool used by this instance.
+ */
+- (GSThreadPool*) threadPool;
+
 /**
  * Retrieves additional user information (previously set using the
  * -setUserInfo:forRequest: method) from a request.
@@ -1147,10 +1183,31 @@
  * actually what it seems.
  */
 @interface      WebServerResponse : GSMimeDocument
+/** Blocks (for the time interval specified) further incoming requests
+ * from the same source as the one we are responding to.<br />
+ * A ti value more than zero establishes a new blocking.<br />
+ * A ti value of zero cancels any existing blocking.<br />
+ * A ti value of less than zero is ignored and the value returned by
+ * the -[WebServer blockOnAuthenticationFailure] method is used instead.<br />
+ * Subsequent requests from the blocked source will be responded to with
+ * a 429 status code until the blocking expires.
+ */ 
+- (void) block: (NSTimeInterval)ti;
+
 /** Behaves as [WebServer-setFoldHeaders:] but applies only to the headers
  * in the receiver.
  */
 - (void) setFoldHeaders: (BOOL)aFlag;
+
+/** Sets additional information attached to the response which may then be
+ * retrieved with the -userInfo method.
+ */
+- (void) setUserInfo: (NSObject*)info;
+
+/** Returns the obect previously set with the -setUserInfo: method or nil
+ * if nothing has been set.
+ */
+- (NSObject*) userInfo;
 @end
 #endif
 
